/**
 * 
 */
package com.goldencode.p2j.persist;

import java.util.*;
import java.util.logging.*;
import com.goldencode.ast.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.schema.*;
import com.goldencode.p2j.util.*;

/**
 * A shared pool of {@link com.goldencode.p2j.pattern.PatternEngine pattern engines} used for
 * runtime conversion of dynamically defined persistence constructs, such as temp-tables and
 * queries. A shared pool is used, rather than instantiating pattern engines as needed, because
 * these objects are very expensive to create and configure. The instances in the pool are
 * created and configured to their specific conversion task once, then stored for use many times
 * across contexts.
 * <p>
 * A conversion task is performed by invoking the {@link #runTask(ConversionProfile, Aast...)}
 * method, which requires the caller to specify a profile which defines the conversion task to
 * be performed, and to provide one or more target ASTs on which to operate. Any objects stored
 * during the conversion task are available through the {@link Results} object returned.
 */
public final class ConversionPool
{
   /** Logger */
   private static Logger LOG = LogHelper.getLogger(ConversionPool.class);
   
   /** Singleton instance of this class */
   private static ConversionPool instance = null;
   
   /** Map of pattern engine stacks, keyed by conversion profile */
   private Map<ConversionProfile, Stack<PatternEngine>> pool = new HashMap<>();
   
   /**
    * Private constructor.
    */
   private ConversionPool()
   {
   }
   
   /**
    * Initialize the conversion pool by loading all necessary, permanent schemas and creating the
    * initial pattern engine instances.
    * 
    * @throws  ConfigurationException
    *          if there is an error loading schemas or creating pattern engine instances.
    * @throws  IllegalStateException
    *          if this method is called more than once.
    */
   public synchronized static void initialize()
   throws ConfigurationException
   {
      if (instance != null)
      {
         throw new IllegalStateException(
            "Conversion pool must not be initialized more than once");
      }
      
      long start = System.currentTimeMillis();
      
      instance = new ConversionPool();
      
      // load the permanent schemas needed for conversion
      
      SchemaConfig sc = Configuration.getSchemaConfig();
      Iterator<String> dbIter = sc.defaultDatabases();
      List<String> all = new ArrayList<>();
      while (dbIter.hasNext())
      {
         all.add(dbIter.next());
      }
      all.add(sc.getMetadata().getName());
      
      // load the schema dict files into the user's in-memory AST storage
      Map<String, Aast> pristineDbAsts = new HashMap<>();
      XmlFilePlugin fileLoader = new XmlFilePlugin(null, true);
      
      for (String schema : all)
      {
         try
         {
            String schemaPath = sc.getSchemaResource(schema);
            
            if (schemaPath == null)
               throw new ConfigurationException(
                    String.format("Could not load the pristine %s schema *.dict!", schema));
            
            // load the schema
            Aast ast = fileLoader.loadTree(schemaPath);
            pristineDbAsts.put(schemaPath, ast);
            
            // load the .p2o definitions
            String p2oPath = schemaPath.replaceAll("\\.dict", ".p2o");
            
            if (ClassLoader.getSystemClassLoader().getResource(p2oPath) == null)
            {
               throw new ConfigurationException(
                 String.format("Could not load the pristine %s schema *.p2o!", schema));
            }
            
            ast = fileLoader.loadTree(p2oPath);
            pristineDbAsts.put(p2oPath, ast);
         }
         catch (AstException | ConfigurationException exc)
         {
            pristineDbAsts = null;
            
            String msg = String.format("Could not load the pristine %s schema!", schema);
            
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.log(Level.SEVERE, msg, exc);
            }
            
            throw new RuntimeException(msg, exc);
         }
      }
      
      for (String schema : all)
      {
         String schemaPath = sc.getSchemaResource(schema);
         if (schemaPath == null)
            throw new ConfigurationException(
                  String.format("Could not load the pristine %s schema *.dict!", schema));
         
         // load the schema
         if (pristineDbAsts.containsKey(schemaPath))
         {
            Aast schemaAst = pristineDbAsts.get(schemaPath);
            AstManager.get().saveTree(schemaAst, schemaPath, false);
         }
         
         // load the .p2o definitions
         String p2oPath = schemaPath.replaceAll("\\.dict", ".p2o");
         
         if (pristineDbAsts.containsKey(p2oPath))
         {
            Aast p2oAst = pristineDbAsts.get(p2oPath);
            AstManager.get().saveTree(p2oAst, p2oPath, false);
         }
      }
      
      // add pattern engines
      for (ConversionProfile profile : ConversionProfile.values())
      {
         // TODO: use directory-based configuration to initialize number of engines per
         // profile
         instance.addEngine(profile);
      }
      
      if (LOG.isLoggable(Level.INFO))
      {
         long elapsed = System.currentTimeMillis() - start;
         LOG.log(Level.INFO, "Runtime conversion pool initialized in " + elapsed + " ms.");
      }
   }
   
   /**
    * Run a specific conversion task against the given ASTs, as defined by the provided profile.
    * 
    * @param   profile
    *          Enumeration value defining the conversion profile to apply to the target ASTs.
    * @param   targetAsts
    *          Source ASTs upon which to operate.
    * 
    * @return  An object which contains any objects stored in the pattern engine as a result of
    *          the conversion task.
    */
   static Results runTask(ConversionProfile profile, Aast... targetAsts)
   {
      PatternEngine engine = null;
      
      try
      {
         // get a pattern engine from the pool
         engine = instance.popEngine(profile);
         
         // add the target ASTs
         for (Aast ast : targetAsts)
         {
            engine.addTargetAst(ast);
         }
         
         // run the conversion
         engine.run();
         
         // collect and return results
         return (new Results(engine.getAllStoredObjects()));
      }
      finally
      {
         engine.clearStoredObjects();
         
         // return the pattern engine to the pool
         if (engine != null)
         {
            instance.pushEngine(profile, engine);
         }
      }
   }
   
   /**
    * Add a pattern engine instance to the stack associated with the given conversion profile.
    * 
    * @param   profile
    *          Conversion profile which will determine the task performed by the new pattern
    *          engine instance.
    * @throws  ConfigurationException
    *          if there is an error instantiating and initializing the new pattern engine object.
    */
   private void addEngine(ConversionProfile profile)
   throws ConfigurationException
   {
      Stack<PatternEngine> stack;
      synchronized (pool)
      {
         stack = pool.get(profile);
         if (stack == null)
         {
            stack = new Stack<>();
         }
         pool.put(profile, stack);
      }
      
      synchronized (stack)
      {
         PatternEngine engine = new PatternEngine(profile.getName(), profile.isConditional());
         stack.push(engine);
         stack.notify();
      }
   }
   
   /**
    * Push a pattern engine instance back onto its stack, so it can be used again.
    * 
    * @param   profile
    *          Profile which defines the pattern engine's conversion task.
    * @param   engine
    *          Pattern engine instance to be pushed.
    */
   private void pushEngine(ConversionProfile profile, PatternEngine engine)
   {
      Stack<PatternEngine> stack;
      synchronized (pool)
      {
         stack = pool.get(profile);
      }
      
      synchronized (stack)
      {
         stack.push(engine);
         stack.notify();
      }
   }
   
   /**
    * Pop a pattern engine instance from its stack in order to perform a conversion task.
    * 
    * @param   profile
    *          Profile which defines the pattern engine's conversion task.
    * 
    * @return  Pattern engine instance.
    */
   private PatternEngine popEngine(ConversionProfile profile)
   {
      Stack<PatternEngine> stack;
      synchronized (pool)
      {
         stack = pool.get(profile);
      }
      
      PatternEngine engine;
      synchronized (stack)
      {
         while (stack.isEmpty())
         {
            try
            {
               stack.wait();
            }
            catch (InterruptedException exc)
            {
               // TODO: check whether spurious or real termination
            }
         }
         
         engine = stack.pop();
      }
      
      return engine;
   }
   
   /**
    * Object which holds the objects stored in the pattern engine by its conversion task.
    */
   static class Results
   {
      /** Map of stored objects */
      private final Map<PatternEngine.StorageKey, Object> storedObjects;
      
      /**
       * Constructor.
       * 
       * @param   storedObjects
       *          Map of stored objects.
       */
      private Results(Map<PatternEngine.StorageKey, Object> storedObjects)
      {
         this.storedObjects = new HashMap<>(storedObjects);
      }
      
      /**
       * Get the object stored under the given id.
       * 
       * @param   id
       *          Object id.
       * 
       * @return  Stored object or <code>null</code> if none is found.
       * 
       * @throws  IllegalStateException
       *          if method is invoked before task has run.
       */
      Object getStoredObject(String id)
      {
         return getStoredObject(null, id);
      }
      
      /**
       * Get the object stored under the given category and id.
       * 
       * @param   category
       *          Object category.
       * @param   id
       *          Object id.
       * 
       * @return  Stored object or <code>null</code> if none is found.
       * 
       * @throws  IllegalStateException
       *          if method is invoked before task has run.
       */
      Object getStoredObject(String category, String id)
      {
         if (storedObjects == null)
         {
            throw new IllegalStateException("Cannot retrieve stored object until task has run");
         }
         
         return storedObjects.get(new PatternEngine.StorageKey(category, id));
      }
   }
}
